@@ -1,10 +1,9 @@
 /*
 SMS Server Tools 3
-Copyright (C) Keijo Kasvi
+Copyright (C) 2006- Keijo Kasvi
 http://smstools3.kekekasvi.com/
 
-Based on SMS Server Tools 2 from Stefan Frings
-http://www.meinemullemaus.de/
+Based on SMS Server Tools 2, http://stefanfrings.de/smstools/
 SMS Server Tools version 2 and below are Copyright (C) Stefan Frings.
 
 This program is free software unless you got it under another license directly
@@ -23,6 +22,7 @@ Either version 2 of the License, or (at your option) any later version.
 #include "smsd_cfg.h"
 #include "logging.h"
 #include "charset.h" // required for conversions of partial text content.
+#include "extras.h"
 
 #define MAX_ADDRESS_LENGTH 50
 #define MAX_SMSC_ADDRESS_LENGTH 30
@@ -48,7 +48,7 @@ int add_warning(char *buffer, char *format, ...)
     else
     {
       result = 0;
-      writelogfile(LOG_ERR, process_title, "PDU %s%s", title, text);
+      writelogfile(LOG_ERR, 1, "PDU %s%s", title, text);
     }
   }
 
@@ -169,7 +169,7 @@ int text2pdu(char* text, int length, char* pdu, char* udh)
   if (length>maxsms_pdu-udh_size_septets)
     length=maxsms_pdu-udh_size_septets;
   //clear the tmp buffer
-  for (character=0;character<sizeof(tmp);character++)
+  for (character=0;(size_t)character<sizeof(tmp);character++)
     tmp[character]=0;
   // Convert 8bit text stream to 7bit stream
   for (character=0;character<length;character++)
@@ -215,35 +215,133 @@ void binary2pdu(char* binary, int length, char* pdu)
   }
 }
 
+int set_numberformat(int *numberformat, char *number, int number_type)
+{
+  int nf = NF_INTERNATIONAL;
+  char *p;
+  int nf_national = NF_NATIONAL; // 3.1.16beta2.
+
+  if (DEVICE.national_toa_unknown == 1)
+    nf_national = NF_UNKNOWN;
+
+  if (numberformat)
+    nf = *numberformat;
+
+  if (nf == NF_INTERNATIONAL)
+  {
+    // If international prefixes are defined, international format is used if a number matches to the list.
+    // If it does not match, national format will be used.
+    if (*international_prefixes)
+    {
+      p = international_prefixes;
+      do
+      {
+        if (!strncmp(number, p, strlen(p)))
+          break;
+        p += strlen(p) +1;
+      }
+      while (*p);
+
+      if (!(*p))
+        nf = nf_national;
+    }
+    
+    if (nf == NF_INTERNATIONAL)
+    {
+      // If national prefixes are defined, national format is used if a number matches to the list.
+      if (*national_prefixes)
+      {
+        p = national_prefixes;
+        do
+        {
+          if (!strncmp(number, p, strlen(p)))
+          {
+            nf = nf_national;
+            break;
+          }
+          p += strlen(p) +1;
+        }
+        while (*p);
+      }
+    }
+  }
+
+  // Finally, user setting overrides all...
+  switch (number_type)
+  {
+    case 0: // "unknown"
+      nf = NF_UNKNOWN; // 1000 0001
+      break;
+
+    case 1: // "international"
+      nf = NF_INTERNATIONAL; // 1001 0001
+      break;
+
+    case 2: // "national"
+      nf = NF_NATIONAL; // 1010 0001
+      break;
+  }
+
+  if (numberformat)
+    *numberformat = nf;
+
+  return nf;
+}
+
 // Make the PDU string from a mesage text and destination phone number.
 // The destination variable pdu has to be big enough. 
 // alphabet indicates the character set of the message.
 // flash_sms enables the flash flag.
 // mode select the pdu version (old or new).
 // if udh is true, then udh_data contains the optional user data header in hex dump, example: "05 00 03 AF 02 01"
-
-void make_pdu(char* number, char* message, int messagelen, int alphabet, int flash_sms, int report, int with_udh,
-              char* udh_data, char* mode, char* pdu, int validity, int replace_msg, int system_msg)
+void make_pdu(
+  char *number,
+  char *message,
+  int messagelen,
+  int alphabet,
+  int flash_sms,
+  int report,
+  int with_udh,
+  char *udh_data,
+  char *mode,
+  char *pdu,
+  int validity,
+  int replace_msg,
+  int system_msg,
+  int number_type,
+  char *smsc,
+  // 3.1.16beta2:
+  int message_reference,
+  int reject_duplicates,
+  int reply_path,
+  int sms_class,
+  int tp_dcs,
+  int ping
+)
 {
   int coding;
   int flags;
-  char tmp[50];
+  char tmp[SIZE_TO];
   char tmp2[500];
   int numberformat;
   int numberlength;
   char *p;
   int l;
+  char tmp_smsc[SIZE_SMSC];
 
-  if (number[0]=='s')  // Is number starts with s, then send it without number format indicator
+  if (number[0] == 's')  // Is number starts with s, then send it without number format indicator
   {
-    numberformat=129; // 1000 0001
-    strcpy(tmp,number+1);
+    numberformat = NF_UNKNOWN;
+    snprintf(tmp, sizeof(tmp), "%s", number +1);
   }
   else
   {
-    numberformat=145; // 1001 0001
-    strcpy(tmp,number);
+    numberformat = NF_INTERNATIONAL;
+    snprintf(tmp, sizeof(tmp), "%s", number);
   }
+
+  set_numberformat(&numberformat, tmp, number_type);
+
   numberlength=strlen(tmp);
   // terminate the number with F if the length is odd
   if (numberlength%2)
@@ -251,6 +349,19 @@ void make_pdu(char* number, char* message, int messagelen, int alphabet, int fla
   // Swap every second character
   swapchars(tmp);
 
+  // 3.1.12:
+  *tmp_smsc = 0;
+  if (DEVICE.smsc_pdu && (smsc[0] || DEVICE.smsc[0]))
+  {
+    p = (smsc[0]) ? smsc : DEVICE.smsc;
+    while (*p == '+')
+      p++;
+    snprintf(tmp_smsc, sizeof(tmp_smsc), "%s", p);
+    if (strlen(tmp_smsc) % 2 && strlen(tmp_smsc) < sizeof(tmp_smsc) - 1)
+      strcat(tmp_smsc, "F");
+    swapchars(tmp_smsc);
+  }
+
   flags=1; // SMS-Sumbit MS to SMSC
   if (with_udh)
     flags+=64; // User Data Header
@@ -258,18 +369,50 @@ void make_pdu(char* number, char* message, int messagelen, int alphabet, int fla
     flags+=16; // Validity field
   if (report>0)
     flags+=32; // Request Status Report
+  if (reject_duplicates)
+    flags += 4;
+  if (reply_path)
+    flags += 128;
 
-  if (alphabet == 1)
+  if (alphabet == ALPHABET_BINARY)
     coding = 4; // 8bit binary
-  else if (alphabet == 2)
+  else if (alphabet == ALPHABET_UCS2)
     coding = 8; // 16bit
   else
     coding = 0; // 7bit
+
   if (flash_sms > 0)
     coding += 0x10; // Bits 1 and 0 have a message class meaning (class 0, alert)
+  else
+  {
+    // 3.1.16beta2. Invalid values for class are just ignored.
+    // Note: system_msg will override "Alphabet", "Flash" and "Class" settings.
+    switch (sms_class)
+    {
+      case 0:                  // Immediate display (alert)
+        coding += 0x10;
+        break;
+
+      case 1:                  // ME specific
+        coding += 0x11;
+        break;
+
+      case 2:                  // SIM specific
+        coding += 0x12;
+        break;
+
+      case 3:                  // TE specific
+        coding += 0x13;
+        break;
+    }
+  }
+
+  // 3.1.16beta2: tp_dcs overrides all but system_msg:
+  if (tp_dcs >= 0 && tp_dcs <= 0xFF)
+    coding = tp_dcs;
 
   /* Create the PDU string of the message */
-  if (alphabet==1 || alphabet==2 || system_msg)
+  if (alphabet == ALPHABET_BINARY || alphabet == ALPHABET_UCS2 || system_msg)
   {
     // Unicode message can be concatenated:
     //if (alphabet == 2 && with_udh)
@@ -278,7 +421,7 @@ void make_pdu(char* number, char* message, int messagelen, int alphabet, int fla
     {
       strcpy(tmp2, udh_data);
       while ((p = strchr(tmp2, ' ')))
-        strcpy(p, p +1);
+        strcpyo(p, p +1);
       l = strlen(tmp2) /2;
       binary2pdu(message, messagelen, strchr(tmp2, 0));
       messagelen += l;
@@ -288,18 +431,48 @@ void make_pdu(char* number, char* message, int messagelen, int alphabet, int fla
   }
   else
     messagelen=text2pdu(message,messagelen,tmp2,udh_data);
+
   /* concatenate the first part of the PDU string */
   if (strcmp(mode,"old")==0)
     sprintf(pdu,"%02X00%02X%02X%s00%02X%02X",flags,numberlength,numberformat,tmp,coding,messagelen);
   else
   {
+    int proto = 0;
+
     if (validity < 0 || validity > 255)
       validity = validity_period;
-    sprintf(pdu, "00%02X00%02X%02X%s%02X%02X%02X%02X",
-            flags, numberlength, numberformat, tmp,
-            (system_msg)? 0x40 : (replace_msg >= 1 && replace_msg <= 7)? 0x40 +replace_msg : 0,
-            (system_msg)? 0xF4 : coding, validity, messagelen);
+
+    if (system_msg)
+    {
+      proto = 0x40;
+      coding = 0xF4;  // binary
+
+      // 3.1.7:
+      if (system_msg == 2)
+      {
+        proto += (0x7F - 0x40); // SS (no show)
+        coding += 2;    // store to sim
+      }
+    }
+    else if (replace_msg >= 1 && replace_msg <= 7)
+      proto = 0x40 + replace_msg;
+    else if (ping)
+      proto = 0x40;
+
+    // 3.1.12:
+    //sprintf(pdu, "00%02X00%02X%02X%s%02X%02X%02X%02X", flags, numberlength, numberformat, tmp, proto, coding, validity, messagelen);
+    if (*tmp_smsc)
+      sprintf(pdu, "%02X%s%s", (int)strlen(tmp_smsc) / 2 + 1, (tmp_smsc[1] == '0')? "81": "91", tmp_smsc);
+    else
+      strcpy(pdu, "00");
+
+    if (message_reference < 0 || message_reference > 255)
+      message_reference = 0;
+
+    sprintf(strchr(pdu, 0), "%02X%02X%02X%02X%s%02X%02X%02X%02X", flags, message_reference,
+            numberlength, numberformat, tmp, proto, coding, validity, messagelen);
   }
+
   /* concatenate the text to the PDU string */
   strcat(pdu,tmp2);
 }
@@ -458,12 +631,12 @@ int explain_udh(char *udh_type, char *pdu)
     return -1;
   strcpy(buffer, pdu);
   while ((p = strchr(buffer, ' ')))
-    strcpy(p, p +1);
+    strcpyo(p, p +1);
 
   if ((udh_length = octet2bin_check(buffer)) < 0)
     return -1;
   udh_length++;
-  if (udh_length *2 > strlen(buffer))
+  if ((size_t)(udh_length *2) > strlen(buffer))
     return -1;
   sprintf(udh_type, "Length=%i", udh_length);
   idx = 1;
@@ -510,13 +683,15 @@ int explain_udh(char *udh_type, char *pdu)
       case 0x21: p = "Hyperlink format element"; break;
       case 0x22: p = "Reply Address Element"; break;
       case 0x23: p = "Enhanced Voice Mail Information"; break;
+      case 0x24: p = "Single Shift Character Set information"; break;
+      case 0x25: p = "Locking Shift Character Set information"; break;
     }
 
     if (!p)
     {
       if (i >= 0x1B && i <= 0x1F)
         p = "Reserved for future EMS features";
-      else if (i >= 0x24 && i <= 0x6F)
+      else if (i >= 0x26 && i <= 0x6F)
         p = "Reserved for future use";
       else if (i >= 0x70 && i <= 0x7F)
         p = "(U)SIM Toolkit Security Headers";
@@ -540,7 +715,7 @@ int explain_udh(char *udh_type, char *pdu)
     // Next octet is length of data:
     if ((i = octet2bin_check(Src_Pointer +2)) < 0)
       return -1;
-    if (i *2 > strlen(Src_Pointer +4))
+    if ((size_t)(i *2) > strlen(Src_Pointer +4))
       return -1;
     idx += i +2;
     if (idx > udh_length)
@@ -569,7 +744,7 @@ int pdu2text(char *pdu, char *text, int *text_length, int *expected_length,
   int octetcounter;
   int skip_characters = 0;
   char c;
-  char binary;
+  char binary = 0;
   int i;
   int result;
 
@@ -807,7 +982,7 @@ int pdu2binary(char* pdu, char* binary, int *data_length, int *expected_length,
 int explain_toa(char *dest, char *octet_char, int octet_int)
 {
   int result;
-  char *p;
+  char *p = "reserved";
 
   if (octet_char)
     result = octet2bin_check(octet_char);
@@ -825,7 +1000,7 @@ int explain_toa(char *dest, char *octet_char, int octet_int)
       case 4: p = "subsciber"; break;
       case 5: p = "alphanumeric"; break;
       case 6: p = "abbreviated"; break;
-      case 7: p = "reserved"; break;
+      //case 7: p = "reserved"; break;
     }
 
     if (octet_char)
@@ -842,7 +1017,7 @@ int explain_toa(char *dest, char *octet_char, int octet_int)
       case 8: p = "national"; break;
       case 9: p = "private"; break;
       case 10: p = "ERMES"; break;
-      default: p = "reserved"; break;
+      //default: p = "reserved"; break;
     }
     sprintf(strchr(dest, 0), ", %s", p);
   }
@@ -850,6 +1025,63 @@ int explain_toa(char *dest, char *octet_char, int octet_int)
   return result;
 }
 
+// 3.1.14:
+void explain_status(char *dest, size_t size_dest, int status)
+{
+  char *p = "unknown";
+
+  switch (status)
+  {
+    case 0: p = "Ok,short message received by the SME"; break;
+    case 1: p = "Ok,short message forwarded by the SC to the SME but the SC is unable to confirm delivery"; break;
+    case 2: p = "Ok,short message replaced by the SC"; break;
+
+    // Temporary error, SC still trying to transfer SM
+    case 32: p = "Still trying,congestion"; break;
+    case 33: p = "Still trying,SME busy"; break;
+    case 34: p = "Still trying,no response sendr SME"; break;
+    case 35: p = "Still trying,service rejected"; break;
+    case 36: p = "Still trying,quality of service not available"; break;
+    case 37: p = "Still trying,error in SME"; break;
+    // 38...47: Reserved
+    // 48...63: Values specific to each SC
+
+    // Permanent error, SC is not making any more transfer attempts
+    case 64: p = "Error,remote procedure error"; break;
+    case 65: p = "Error,incompatible destination"; break;
+    case 66: p = "Error,connection rejected by SME"; break;
+    case 67: p = "Error,not obtainable"; break;
+    case 68: p = "Error,quality of service not available"; break;
+    case 69: p = "Error,no interworking available"; break;
+    case 70: p = "Error,SM validity period expired"; break;
+    case 71: p = "Error,SM deleted by originating SME"; break;
+    case 72: p = "Error,SM deleted by SC administration"; break;
+    case 73: p = "Error,SM does not exist"; break;
+    // 74...79: Reserved
+    // 80...95: Values specific to each SC
+
+    // Permanent error, SC is not making any more transfer attempts
+    case 96: p = "Error,congestion"; break;
+    case 97: p = "Error,SME busy"; break;
+    case 98: p = "Error,no response sendr SME"; break;
+    case 99: p = "Error,service rejected"; break;
+    case 100: p = "Error,quality of service not available"; break;
+    case 101: p = "Error,error in SME"; break;
+    // 102...105: Reserved
+    // 106...111: Reserved
+    // 112...127: Values specific to each SC
+    // 128...255: Reserved
+
+    default:
+      if (status >= 48 && status <= 63)
+        p = "Temporary error, SC specific, unknown";
+      else if ((status >= 80 && status <= 95) || (status >= 112 && status <= 127))
+        p = "Permanent error, SC specific, unknown";
+  }
+
+  snprintf(dest, size_dest, "%s", p);
+}
+
 // Subroutine for messages type 0 (SMS-Deliver)
 // Input:
 // Src_Pointer points to the PDU string
@@ -857,13 +1089,15 @@ int explain_toa(char *dest, char *octet_char, int octet_int)
 // sendr Sender
 // date and time Date/Time-stamp
 // message the message text or binary data
+// bin_udh: 1 if udh is taken from the PDU with binary messages too
 // returns length of message
 int split_type_0(char *full_pdu, char* Src_Pointer, int* alphabet, char* sendr, char* date, char* time,
                  char* message, int *message_length, int *expected_length, int with_udh, char* udh_data,
-                 char *udh_type, char *from_toa, int *replace, char **err_str, char *warning_headers, int *flash)
+                 char *udh_type, char *from_toa, int *replace, char **err_str, char *warning_headers,
+                 int *flash, int bin_udh)
 {
   int Length;
-  int padding;
+  int padding = 0;
   char tmpsender[100];
   int result = 0;
   int i;
@@ -879,9 +1113,13 @@ int split_type_0(char *full_pdu, char* Src_Pointer, int* alphabet, char* sendr,
   else
   {
     Length = octet2bin_check(Src_Pointer);
-    if (Length < 1 || Length > MAX_ADDRESS_LENGTH)
+    // 3.1.5: Sender address length can be zero. There is still address type octet (80).
+    if (Length < 0 || Length > MAX_ADDRESS_LENGTH)
       pdu_error(err_str, 0, Src_Pointer -full_pdu, 2, "Invalid sender address length: \"%.2s\"", Src_Pointer);
     else
+    if (Length == 0)
+      Src_Pointer += 4;
+    else
     {
       padding=Length%2;
       Src_Pointer+=2;
@@ -895,7 +1133,7 @@ int split_type_0(char *full_pdu, char* Src_Pointer, int* alphabet, char* sendr,
         Src_Pointer += 2;
         if ((i & 112) == 80)
         {  // Sender is alphanumeric
-          if (strlen(Src_Pointer) < Length +padding)
+          if (strlen(Src_Pointer) < (size_t)(Length +padding))
             pdu_error(err_str, 0, Src_Pointer -full_pdu, Length +padding,
                       "While trying to read sender address (alphanumeric, length %i): %s",
                       Length +padding, err_too_short);
@@ -909,7 +1147,7 @@ int split_type_0(char *full_pdu, char* Src_Pointer, int* alphabet, char* sendr,
         }
         else
         {  // sender is numeric
-          if (strlen(Src_Pointer) < Length +padding)
+          if (strlen(Src_Pointer) < (size_t)(Length +padding))
             pdu_error(err_str, 0, Src_Pointer -full_pdu, Length +padding,
                       "While trying to read sender address (numeric, length %i): %s",
                       Length +padding, err_too_short);
@@ -936,11 +1174,11 @@ int split_type_0(char *full_pdu, char* Src_Pointer, int* alphabet, char* sendr,
             }
 
             for (i = 0; sendr[i]; i++)
-              if (!isdigit(sendr[i]))
+              if (!isdigitc(sendr[i]))
               {
                 // Not a fatal error (?)
                 //pdu_error(err_str, 0, Src_Pointer -full_pdu, Length +padding, "Invalid character(s) in sender address: \"%s\"", sendr);
-                //*sendr = 0;
+                // *sendr = 0;
                 add_warning(warning_headers, "Invalid character(s) in sender address.");
                 break;
               }
@@ -990,7 +1228,7 @@ int split_type_0(char *full_pdu, char* Src_Pointer, int* alphabet, char* sendr,
             {
               Src_Pointer += 2;
               sprintf(date,"%c%c-%c%c-%c%c",Src_Pointer[1],Src_Pointer[0],Src_Pointer[3],Src_Pointer[2],Src_Pointer[5],Src_Pointer[4]);
-              if (!isdigit(date[0]) || !isdigit(date[1]) || !isdigit(date[3]) || !isdigit(date[4]) || !isdigit(date[6]) || !isdigit(date[7]))
+              if (!isdigitc(date[0]) || !isdigitc(date[1]) || !isdigitc(date[3]) || !isdigitc(date[4]) || !isdigitc(date[6]) || !isdigitc(date[7]))
               {
                 pdu_error(err_str, 0, Src_Pointer -full_pdu, 6, "Invalid character(s) in date of Service Centre Time Stamp: \"%s\"", date);
                 *date = 0;
@@ -999,13 +1237,13 @@ int split_type_0(char *full_pdu, char* Src_Pointer, int* alphabet, char* sendr,
               {
                 // Not a fatal error (?)
                 //pdu_error(err_str, 0, Src_Pointer -full_pdu, 6, "Invalid value(s) in date of Service Centre Time Stamp: \"%s\"", date);
-                //*date = 0;
+                // *date = 0;
                 add_warning(warning_headers, "Invalid values(s) in date of Service Centre Time Stamp.");
               }
 
               Src_Pointer += 6;
               sprintf(time,"%c%c:%c%c:%c%c",Src_Pointer[1],Src_Pointer[0],Src_Pointer[3],Src_Pointer[2],Src_Pointer[5],Src_Pointer[4]);
-              if (!isdigit(time[0]) || !isdigit(time[1]) || !isdigit(time[3]) || !isdigit(time[4]) || !isdigit(time[6]) || !isdigit(time[7]))
+              if (!isdigitc(time[0]) || !isdigitc(time[1]) || !isdigitc(time[3]) || !isdigitc(time[4]) || !isdigitc(time[6]) || !isdigitc(time[7]))
               {
                 pdu_error(err_str, 0, Src_Pointer -full_pdu, 6, "Invalid character(s) in time of Service Centre Time Stamp: \"%s\"", time);
                 *time = 0;
@@ -1014,7 +1252,7 @@ int split_type_0(char *full_pdu, char* Src_Pointer, int* alphabet, char* sendr,
               {
                 // Not a fatal error (?)
                 //pdu_error(err_str, 0, Src_Pointer -full_pdu, 6, "Invalid value(s) in time of Service Centre Time Stamp: \"%s\"", time);
-                //*time = 0;
+                // *time = 0;
                 add_warning(warning_headers, "Invalid values(s) in time of Service Centre Time Stamp.");
               }
 
@@ -1049,8 +1287,10 @@ int split_type_0(char *full_pdu, char* Src_Pointer, int* alphabet, char* sendr,
       {
         // With binary messages udh is NOT taken from the PDU.
         i = with_udh;
-        if (*alphabet == 1)
-          i = 0;
+        // 3.1.5: it should work now:
+        if (bin_udh == 0)
+          if (*alphabet == 1)
+            i = 0;
         if ((result = pdu2binary(Src_Pointer, message, message_length, expected_length, i, udh_data, udh_type, &errorpos)) < 0)
           pdu_error(err_str, 0, Src_Pointer -full_pdu +errorpos, 0, "While reading TP-UD (%s): %s",
                     (*alphabet == 1)? "binary" : "UCS2 text",
@@ -1079,7 +1319,6 @@ int split_type_2(char *full_pdu, char* Src_Pointer,char* sendr, char* date,char*
   char tmpsender[100];
   int messageid;
   int i;
-  char *p;
   const char SR_MessageId[] = "Message_id:"; // Fixed title inside the status report body.
   const char SR_Status[] = "Status:"; // Fixed title inside the status report body.
 
@@ -1116,7 +1355,7 @@ int split_type_2(char *full_pdu, char* Src_Pointer,char* sendr, char* date,char*
           Src_Pointer += 2;
           if ((i & 112) == 80)
           {  // Sender is alphanumeric
-            if (strlen(Src_Pointer) < Length +padding)
+            if (strlen(Src_Pointer) < (size_t)(Length +padding))
               pdu_error(err_str, 0, Src_Pointer -full_pdu, Length +padding,
                         "While trying to read recipient address (alphanumeric, length %i): %s",
                         Length +padding, err_too_short);
@@ -1130,7 +1369,7 @@ int split_type_2(char *full_pdu, char* Src_Pointer,char* sendr, char* date,char*
           }
           else
           {  // sender is numeric
-            if (strlen(Src_Pointer) < Length +padding)
+            if (strlen(Src_Pointer) < (size_t)(Length +padding))
               pdu_error(err_str, 0, Src_Pointer -full_pdu, Length +padding,
                         "While trying to read recipient address (numeric, length %i): %s",
                         Length +padding, err_too_short);
@@ -1157,11 +1396,11 @@ int split_type_2(char *full_pdu, char* Src_Pointer,char* sendr, char* date,char*
               }
 
               for (i = 0; sendr[i]; i++)
-                if (!isdigit(sendr[i]))
+                if (!isdigitc(sendr[i]))
                 {
                   // Not a fatal error (?)
                   //pdu_error(err_str, 0, Src_Pointer -full_pdu, Length +padding, "Invalid character(s) in recipient address: \"%s\"", sendr);
-                  //*sendr = 0;
+                  // *sendr = 0;
                   add_warning(warning_headers, "Invalid character(s) in recipient address.");
                   break;
                 }
@@ -1177,7 +1416,7 @@ int split_type_2(char *full_pdu, char* Src_Pointer,char* sendr, char* date,char*
             {
               // get SMSC timestamp
               sprintf(date,"%c%c-%c%c-%c%c",Src_Pointer[1],Src_Pointer[0],Src_Pointer[3],Src_Pointer[2],Src_Pointer[5],Src_Pointer[4]);
-              if (!isdigit(date[0]) || !isdigit(date[1]) || !isdigit(date[3]) || !isdigit(date[4]) || !isdigit(date[6]) || !isdigit(date[7]))
+              if (!isdigitc(date[0]) || !isdigitc(date[1]) || !isdigitc(date[3]) || !isdigitc(date[4]) || !isdigitc(date[6]) || !isdigitc(date[7]))
               {
                 pdu_error(err_str, 0, Src_Pointer -full_pdu, 6, "Invalid character(s) in date of SMSC Timestamp: \"%s\"", date);
                 *date = 0;
@@ -1186,13 +1425,13 @@ int split_type_2(char *full_pdu, char* Src_Pointer,char* sendr, char* date,char*
               {
                 // Not a fatal error (?)
                 //pdu_error(err_str, 0, Src_Pointer -full_pdu, 6, "Invalid value(s) in date of SMSC Timestamp: \"%s\"", date);
-                //*date = 0;
+                // *date = 0;
                 add_warning(warning_headers, "Invalid value(s) in date of SMSC Timestamp.");
               }
 
               Src_Pointer += 6;
               sprintf(time,"%c%c:%c%c:%c%c",Src_Pointer[1],Src_Pointer[0],Src_Pointer[3],Src_Pointer[2],Src_Pointer[5],Src_Pointer[4]);
-              if (!isdigit(time[0]) || !isdigit(time[1]) || !isdigit(time[3]) || !isdigit(time[4]) || !isdigit(time[6]) || !isdigit(time[7]))
+              if (!isdigitc(time[0]) || !isdigitc(time[1]) || !isdigitc(time[3]) || !isdigitc(time[4]) || !isdigitc(time[6]) || !isdigitc(time[7]))
               {
                 pdu_error(err_str, 0, Src_Pointer -full_pdu, 6, "Invalid character(s) in time of SMSC Timestamp: \"%s\"", time);
                 *time = 0;
@@ -1201,7 +1440,7 @@ int split_type_2(char *full_pdu, char* Src_Pointer,char* sendr, char* date,char*
               {
                 // Not a fatal error (?)
                 //pdu_error(err_str, 0, Src_Pointer -full_pdu, 6, "Invalid value(s) in time of SMSC Timestamp: \"%s\"", time);
-                //*time = 0;
+                // *time = 0;
                 add_warning(warning_headers, "Invalid value(s) in time of SMSC Timestamp.");
               }
 
@@ -1226,8 +1465,8 @@ int split_type_2(char *full_pdu, char* Src_Pointer,char* sendr, char* date,char*
             {
               // get Discharge timestamp
               sprintf(temp,"%c%c-%c%c-%c%c %c%c:%c%c:%c%c",Src_Pointer[1],Src_Pointer[0],Src_Pointer[3],Src_Pointer[2],Src_Pointer[5],Src_Pointer[4],Src_Pointer[7],Src_Pointer[6],Src_Pointer[9],Src_Pointer[8],Src_Pointer[11],Src_Pointer[10]);
-              if (!isdigit(temp[0]) || !isdigit(temp[1]) || !isdigit(temp[3]) || !isdigit(temp[4]) || !isdigit(temp[6]) || !isdigit(temp[7]) || 
-                  !isdigit(temp[9]) || !isdigit(temp[10]) || !isdigit(temp[12]) || !isdigit(temp[13]) || !isdigit(temp[15]) || !isdigit(temp[16]))
+              if (!isdigitc(temp[0]) || !isdigitc(temp[1]) || !isdigitc(temp[3]) || !isdigitc(temp[4]) || !isdigitc(temp[6]) || !isdigitc(temp[7]) || 
+                  !isdigitc(temp[9]) || !isdigitc(temp[10]) || !isdigitc(temp[12]) || !isdigitc(temp[13]) || !isdigitc(temp[15]) || !isdigitc(temp[16]))
                 pdu_error(err_str, 0, Src_Pointer -full_pdu, 12, "Invalid character(s) in Discharge Timestamp: \"%s\"", temp);
               else if (atoi(temp +3) > 12 || atoi(temp +6) > 31 || atoi(temp +9) > 24 || atoi(temp +12) > 59 || atoi(temp +16) > 59)
               {
@@ -1250,7 +1489,13 @@ int split_type_2(char *full_pdu, char* Src_Pointer,char* sendr, char* date,char*
 
             if (!(*err_str))
             {
-              sprintf(strchr(result, 0), "Discharge_timestamp: %s", temp);
+              char buffer[128];
+
+              // 3.1.20: If custom datetime_format is defined, use it for Discharge_timestamp:
+              temp[8] = '\0';
+              make_datetime_string(buffer, sizeof(buffer), temp, temp +9, 0);
+              sprintf(strchr(result, 0), "Discharge_timestamp: %s", buffer);
+
               if (strlen(Src_Pointer) < 2)
                 pdu_error(err_str, 0, Src_Pointer -full_pdu, 2, "While trying to read Status octet: %s", err_too_short);
               else
@@ -1260,40 +1505,8 @@ int split_type_2(char *full_pdu, char* Src_Pointer,char* sendr, char* date,char*
                   pdu_error(err_str, 0, Src_Pointer -full_pdu, 2, "Invalid Status octet: \"%.2s\"", Src_Pointer);
                 else
                 {
-                  switch (status)
-                  {
-                    case 0: p = "Ok,short message received by the SME"; break;
-                    case 1: p = "Ok,short message forwarded by the SC to the SME but the SC is unable to confirm delivery"; break;
-                    case 2: p = "Ok,short message replaced by the SC"; break;
-
-                    case 32: p = "Still trying,congestion"; break;
-                    case 33: p = "Still trying,SME busy"; break;
-                    case 34: p = "Still trying,no response sendr SME"; break;
-                    case 35: p = "Still trying,service rejected"; break;
-                    case 36: p = "Still trying,quality of service not available"; break;
-                    case 37: p = "Still trying,error in SME"; break;
-
-                    case 64: p = "Error,remote procedure error"; break;
-                    case 65: p = "Error,incompatible destination"; break;
-                    case 66: p = "Error,connection rejected by SME"; break;
-                    case 67: p = "Error,not obtainable"; break;
-                    case 68: p = "Error,quality of service not available"; break;
-                    case 69: p = "Error,no interworking available"; break;
-                    case 70: p = "Error,SM validity period expired"; break;
-                    case 71: p = "Error,SM deleted by originating SME"; break;
-                    case 72: p = "Error,SM deleted by SC administration"; break;
-                    case 73: p = "Error,SM does not exist"; break;
-
-                    case 96: p = "Error,congestion"; break;
-                    case 97: p = "Error,SME busy"; break;
-                    case 98: p = "Error,no response sendr SME"; break;
-                    case 99: p = "Error,service rejected"; break;
-                    case 100: p = "Error,quality of service not available"; break;
-                    case 101: p = "Error,error in SME"; break;
-
-                    default: p = "unknown";
-                  }
-                  sprintf(strchr(result, 0), "\n%s %i,%s", SR_Status, status, p);
+                  explain_status(buffer, sizeof(buffer), status);
+                  sprintf(strchr(result, 0), "\n%s %i,%s", SR_Status, status, buffer);
                 }
               }
             }
@@ -1322,7 +1535,8 @@ int split_type_2(char *full_pdu, char* Src_Pointer,char* sendr, char* date,char*
 // Returns the length of the message 
 int splitpdu(char *pdu, char *mode, int *alphabet, char *sendr, char *date, char *time, char *message,
              char *smsc, int *with_udh, char *a_udh_data, char *a_udh_type, int *is_statusreport,
-             int *is_unsupported_pdu, char *from_toa, int *report, int *replace, char *warning_headers, int *flash)
+             int *is_unsupported_pdu, char *from_toa, int *report, int *replace, char *warning_headers,
+             int *flash, int bin_udh)
 {
   int Length;
   int Type;
@@ -1346,7 +1560,7 @@ int splitpdu(char *pdu, char *mode, int *alphabet, char *sendr, char *date, char
   // Patch for Wavecom SR memory reading:
   if (strncmp(pdu, "000000FF00", 10) == 0)
   {
-    strcpy(pdu, pdu +8);
+    strcpyo(pdu, pdu +8);
     while (strlen(pdu) < 52)
       strcat(pdu, "00");
   }
@@ -1411,7 +1625,7 @@ int splitpdu(char *pdu, char *mode, int *alphabet, char *sendr, char *date, char
             {
               Length = Length *2 -2;
               // No padding because the given value is number of octets.
-              if (strlen(Pointer) < Length +4)
+              if (strlen(Pointer) < (size_t)(Length +4))
                 pdu_error(&err_str, 0, Pointer -pdu, Length +4, "While trying to read sender SMSC address (length %i): %s",
                           Length, err_too_short);
               else
@@ -1447,11 +1661,11 @@ int splitpdu(char *pdu, char *mode, int *alphabet, char *sendr, char *date, char
                     if (smsc[Length -1] == 'F')
                       smsc[Length -1] = 0;
                     for (i = 0; smsc[i]; i++)
-                      if (!isdigit(smsc[i]))
+                      if (!isdigitc(smsc[i]))
                       {
                         // Not a fatal error (?)
                         //pdu_error(&err_str, 0, Pointer -pdu, Length, "Invalid character(s) in numeric SMSC address: \"%s\"", smsc);
-                        //*smsc = 0;
+                        // *smsc = 0;
                         add_warning(warning_headers, "Invalid character(s) in numeric SMSC address");
                         break;
                       }
@@ -1490,7 +1704,7 @@ int splitpdu(char *pdu, char *mode, int *alphabet, char *sendr, char *date, char
             {
               Pointer += 2;
               result = split_type_0(pdu, Pointer, alphabet, sendr, date, time, message, &message_length, &expected_length,
-                                    *with_udh, udh_data, udh_type, from_toa, replace, &err_str, warning_headers, flash);
+                                    *with_udh, udh_data, udh_type, from_toa, replace, &err_str, warning_headers, flash, bin_udh);
 
               if (err_str && *udh_type)
                 pdu_error(&err_str, "", -1, 0, "Message has Udh_type: %s", udh_type);
@@ -1509,8 +1723,8 @@ int splitpdu(char *pdu, char *mode, int *alphabet, char *sendr, char *date, char
                 else
                 {
                   memcpy(ascii, message, message_length);
-                  ascii[message_length] = 0;
-                  i = decode_ucs2(ascii, message_length);
+                  ucs2utf(ascii, message_length, sizeof(ascii));
+                  i = iso_utf8chars(ascii);
                   expected_length /= 2;
                 }
 
@@ -1592,3 +1806,218 @@ int splitpdu(char *pdu, char *mode, int *alphabet, char *sendr, char *date, char
 
   return result;
 }
+
+int get_pdu_details(char *dest, size_t size_dest, char *pdu, int mnumber)
+{
+  int result = 0;
+  //int udlen;
+  int alphabet;
+  char sender[100];
+  char date[9];
+  char time[9];
+  char ascii[MAXTEXT];
+  char smsc[31];
+  int with_udh;
+  char udh_data[SIZE_UDH_DATA];
+  char udh_type[SIZE_UDH_TYPE];
+  int is_statusreport;
+  int is_unsupported_pdu;
+  char from_toa[51];
+  int report;
+  int replace;
+  char warning_headers[SIZE_WARNING_HEADERS];
+  int flash;
+  int bin_udh = 1;
+  int m_id = 999; // real id for concatenated messages only, others use this.
+  size_t length_sender = 32;
+  int p_count = 1;
+  int p_number = 1;
+  char buffer[100]; // 3.1.6: increased size, was too small (51).
+  int i;
+  char sort_ch;
+  char *p;
+
+  /*udlen = */splitpdu(pdu, DEVICE.mode, &alphabet, sender, date, time, ascii, smsc, &with_udh, udh_data, udh_type,
+                   &is_statusreport, &is_unsupported_pdu, from_toa, &report, &replace, warning_headers, &flash, bin_udh);
+
+  if (is_unsupported_pdu)
+  {
+    writelogfile(LOG_ERR, 1, "Message %i, unsupported PDU.", mnumber);
+    result = 1;
+  }
+  else
+  {
+    if (with_udh)
+    {
+      if (get_remove_concatenation(udh_data, &m_id, &p_count, &p_number) < 0)
+      {
+        writelogfile(LOG_ERR, 1, "Message %i, error while checking UDH_DATA.", mnumber);
+        result = 2;
+      }
+    }
+
+    if (!result)
+    {
+      if (strlen(sender) > length_sender)
+      {
+        writelogfile(LOG_ERR, 1, "Message %i, too long sender field.", mnumber);
+        result = 3;
+      }    
+
+      while (strlen(sender) < length_sender)
+        strcat(sender, " ");
+
+      if (DEVICE.priviledged_numbers[0])
+        p = DEVICE.priviledged_numbers;
+      else
+      if (priviledged_numbers[0])
+        p = priviledged_numbers;
+      else
+        p = 0;
+
+      sort_ch = 'Z';
+      if (p)
+      {
+        i = 0;
+        while (*p)
+        {
+          if (strncmp(sender, p, strlen(p)) == 0)
+          {
+            sort_ch = 'A' +i;
+            break;
+          }
+          i++;
+          p = strchr(p, 0) +1;
+        }
+      }
+/*
+001 A 358401234567____________________111 001/001 r 00-00-00 00-00-00n
+mnumber
+    sort_ch
+      sender
+                                      m_id
+                                          p_number
+                                              p_count
+                                                  incoming / report
+                                                    date
+                                                             time
+
+      snprintf(buffer, sizeof(buffer), "%.03i %c %s%.03i %.03i/%.03i %c %-8.8s %-8.8s\n",
+               mnumber, sort_ch, sender, m_id, p_number, p_count, (is_statusreport) ? 'r' : 'i', date, time);
+
+3.1.7:
+001 A 00-00-00 00-00-00 358401234567____________________111 001/001 rn
+mnumber
+    sort_ch
+      date
+               time
+                        sender
+                                                        m_id
+                                                            p_number
+                                                                p_count
+                                                                    incoming / report
+*/
+      snprintf(buffer, sizeof(buffer), "%.03i %c %-8.8s %-8.8s %s%.03i %.03i/%.03i %c\n", mnumber, sort_ch, date, time, sender, m_id, p_number, p_count, (is_statusreport) ? 'r' : 'i');
+
+      if (strlen(dest) +strlen(buffer) < size_dest)
+        strcat(dest, buffer);
+      else
+      {
+        writelogfile(LOG_ERR, 1, "Message %i, not enough storage space.", mnumber);
+        result = 4;
+      }
+    }
+  }
+
+  return result;
+}
+
+int sort_pdu_helper(const void *a, const void *b)
+{
+  //return(strncmp((char *)a +4, (char *)b +4, 45));
+  return (strncmp((char *) a + 4, (char *) b + 4, 63));
+}
+
+void sort_pdu_details(char *dest)
+{
+  int count;
+
+  count = strlen(dest) / LENGTH_PDU_DETAIL_REC;
+  if (count > 1)
+    qsort((void *)dest, count, LENGTH_PDU_DETAIL_REC, sort_pdu_helper);
+}
+
+int read_pdu_text(char *pdu, size_t pdu_size, char *text)
+{
+  int result = 1;
+  size_t src = 0;
+  size_t dst = 0;
+
+  if (!strncasecmp(pdu, "PDU:", 4))
+    src = 4;
+
+  while (text[src])
+  {
+    if (isXdigit(text[src]))
+    {
+      if (dst >= pdu_size -1)
+      {
+        result = 0;
+        break;
+      }
+
+      pdu[dst++] = text[src];
+    }
+
+    src++;
+  }
+
+  pdu[dst] = '\0';
+
+  return result;
+}
+
+int get_pdu_submit_to(char *to, size_t to_size, char *pdu)
+{
+  int result = 0;
+  char *p = pdu;
+  char *end = pdu + strlen(pdu);
+  int len;
+  char phone[SIZE_TO];
+  int i;
+
+  if ((len = octet2bin_check(p)) >= 0)
+  {
+    for (p += 2; len > 0; len--)
+      p += 2;
+
+    if (p < end && (i = octet2bin_check(p)) > 0 && (i & 0x03) == 0x01)
+    {
+      if ((p += 4) < end && (len = octet2bin_check(p)) >= 2 && (p += 4) < end)
+      {
+        if (len % 2)
+          len++;
+
+        if (len < (int)sizeof(phone))
+        {
+          for (i = 0; i < len; i++)
+            phone[i] = p[i];
+
+          phone[i] = '\0';
+          swapchars(phone);
+
+          if (phone[--i] == 'F')
+            phone[i] = '\0';
+
+          if (strlen(phone) < to_size)
+          {
+            strcpy(to, phone);
+            result = 1;
+          }
+        }
+      }
+    }
+  }
+
+  return result;
+}
